@@ -2,7 +2,7 @@ from __future__ import (absolute_import, division, print_function,
                         unicode_literals)
 
 import collections
-from datetime import datetime, timedelta
+from datetime import datetime, timedelta, time as dtime
 from dateutil.parser import parse as date_parse
 import time as _time
 import threading
@@ -379,92 +379,33 @@ class AlpacaStore(with_metaclass(MetaSingleton, object)):
 
         granularity = self.get_granularity(timeframe, compression)
         if granularity is None:
-            e = AlpacaTimeFrameError()
+            e = AlpacaTimeFrameError('')
             q.put(e.error_response)
             return
+        try:
+            if self.p.usePolygon:
+                cdl = self.get_aggs_from_polygon(dataname,
+                                                 dtbegin,
+                                                 dtend,
+                                                 granularity,
+                                                 compression)
+            else:
+                cdl = self.get_aggs_from_alpaca(dataname,
+                                                dtbegin,
+                                                dtend,
+                                                granularity,
+                                                compression)
+        except AlpacaError as e:
+            print(str(e))
+            q.put(e.error_response)
+            q.put(None)
+            return
+        except Exception as e:
+            print(str(e))
+            q.put({'code': 'error'})
+            q.put(None)
+            return
 
-        dtkwargs = {'start': None, 'end': None}
-        if not dtend:
-            dtend = datetime.utcnow()
-        if not dtbegin:
-            days = 30 if 'd' in granularity else 3
-            delta = timedelta(days=days)
-            dtbegin = dtend - delta
-        dtkwargs['start'] = dtbegin
-        end_dt = None
-        dtkwargs['end'] = dtend
-        end_dt = dtend.isoformat()
-
-        cdl = pd.DataFrame()
-        prevdt = 0
-
-        while True:
-            try:
-                start_dt = None
-                if dtkwargs['start']:
-                    start_dt = dtkwargs['start'].isoformat()
-                if self.p.usePolygon:
-                    response = \
-                        self.oapi.polygon.historic_agg_v2(
-                            dataname,
-                            compression,
-                            granularity,
-                            _from=self.iso_date(start_dt),
-                            to=self.iso_date(end_dt))
-                else:
-                    # response = self.oapi.get_aggs(dataname,
-                    #                               compression,
-                    #                               granularity,
-                    #                               self.iso_date(start_dt),
-                    #                               self.iso_date(end_dt))
-                    # so get_aggs work nicely for days but not for minutes, and
-                    # it is not a documented API. barset on the other hand does
-                    # but we need to manipulate it to be able to work with it
-                    # smoothly and return data the same way polygon does
-                    response = self.oapi.get_barset(
-                        dataname,
-                        granularity,
-                        start=start_dt,
-                        end=end_dt)[dataname]._raw
-                    for bar in response:
-                        # Aggs are in milliseconds, we multiply by 1000 to
-                        # change seconds to ms
-                        bar['t'] *= 1000
-                    response = Aggs({"results": response})
-            except AlpacaError as e:
-                print(str(e))
-                q.put(e.error_response)
-                q.put(None)
-                return
-            except Exception as e:
-                print(str(e))
-                q.put({'code': 'error'})
-                q.put(None)
-                return
-
-            # No result from the server, most likely error
-            if response.df.shape[0] == 0:
-                print(response)
-                q.put({'code': 'error'})
-                q.put(None)
-                return
-            temp = response.df
-            cdl.update(temp)
-            cdl = pd.concat([cdl, temp])
-            cdl = cdl[~cdl.index.duplicated()]
-            prevdt = dtkwargs['start']
-            dtkwargs['start'] = cdl.index[-1].to_pydatetime()
-
-            if prevdt == dtkwargs['start']:  # end of the data
-                break
-
-        freq = str(compression) + ('D' if 'd' in granularity else 'T')
-
-        cdl = cdl.resample(freq).agg({'open': 'first',
-                                      'high': 'max',
-                                      'low': 'min',
-                                      'close': 'last',
-                                      'volume': 'sum'})
         # don't use dt.replace. use localize
         # (https://stackoverflow.com/a/1592837/2739124)
         cdl = cdl.loc[
@@ -477,6 +418,210 @@ class AlpacaStore(with_metaclass(MetaSingleton, object)):
             q.put(r)
         q.put({})  # end of transmission
 
+    def get_aggs_from_polygon(self,
+                              dataname,
+                              dtbegin,
+                              dtend,
+                              granularity,
+                              compression):
+        """
+        so polygon has a much more convenient api for this than alpaca because
+        we could insert the compression in to the api call and we don't need to
+        resample it. but, at this point in time, something is not working
+        properly and data is returned in segments. meaning, we have patches of
+        missing data. e.g we request data from 2020-03-01 to 2020-07-01 and we
+        get something like this: 2020-03-01:2020-03-15, 2020-06-25:2020-07-01
+        so that makes life difficult.. there's no way to know which patch will
+        be returned and which one we should try to get again.
+        so the solution must be, ask data in segments. I select an arbitrary
+        time window of 2 weeks, and split the calls until we get all required
+        data
+        """
+        def _clear_out_of_market_hours(df):
+            """
+            only interested in samples between 9:30, 16:00 NY time
+            """
+            return df.between_time("09:30", "16:00")
+
+        if not dtend:
+            dtend = datetime.utcnow()
+        if not dtbegin:
+            days = 30 if 'd' in granularity else 3
+            delta = timedelta(days=days)
+            dtbegin = dtend - delta
+
+        if granularity == 'day':
+            cdl = self.oapi.polygon.historic_agg_v2(
+                dataname,
+                compression,
+                granularity,
+                _from=self.iso_date(dtbegin.isoformat()),
+                to=self.iso_date(dtend.isoformat())).df
+        else:
+            cdl = pd.DataFrame()
+            segment_start = dtbegin
+            segment_end = segment_start + timedelta(weeks=2) if \
+                dtend - dtbegin >= timedelta(weeks=2) else dtend
+            while segment_end <= dtend:
+                response = self.oapi.polygon.historic_agg_v2(
+                    dataname,
+                    compression,
+                    granularity,
+                    _from=self.iso_date(segment_start.isoformat()),
+                    to=self.iso_date(segment_end.isoformat()))
+                # No result from the server, most likely error
+                if response.df.shape[0] == 0:
+                    raise Exception("received empty response")
+                temp = response.df
+                cdl = pd.concat([cdl, temp])
+                cdl = cdl[~cdl.index.duplicated()]
+                segment_start = segment_end
+                segment_end = segment_start + timedelta(weeks=2) if \
+                    dtend - dtbegin >= timedelta(weeks=2) else dtend
+            cdl = _clear_out_of_market_hours(cdl)
+        return cdl
+
+    def get_aggs_from_alpaca(self,
+                             dataname,
+                             start,
+                             end,
+                             granularity,
+                             compression):
+        """
+        https://alpaca.markets/docs/api-documentation/api-v2/market-data/bars/
+        Alpaca API as a limit of 1000 records per api call. meaning, we need to
+        do multiple calls to get all the required data if the date range is
+        large.
+        also, the alpaca api does not support compression (or, you can't get
+        5 minute bars e.g) so we need to resample the received bars.
+        also, we need to drop out of market records.
+        this function does all of that.
+
+        note:
+        this was the old way of getting the data
+          response = self.oapi.get_aggs(dataname,
+                                        compression,
+                                        granularity,
+                                        self.iso_date(start_dt),
+                                        self.iso_date(end_dt))
+          the thing is get_aggs work nicely for days but not for minutes, and
+          it is not a documented API. barset on the other hand does
+          but we need to manipulate it to be able to work with it
+          smoothly and return data the same way polygon does
+        """
+        def _iterate_api_calls():
+            """
+            you could get max 1000 samples from the server. if we need more
+            than that we need to do several api calls.
+
+            currently the alpaca api supports also 5Min and 15Min so we could
+            optimize server communication time by addressing timeframes
+            """
+            got_all = False
+            curr = end
+            response = []
+            while not got_all:
+                if granularity == 'minute' and compression == 5:
+                    timeframe = "5Min"
+                elif granularity == 'minute' and compression == 15:
+                    timeframe = "15Min"
+                else:
+                    timeframe = granularity
+                r = self.oapi.get_barset(dataname,
+                                         timeframe,
+                                         limit=1000,
+                                         end=curr.isoformat()
+                                         )[dataname]
+                earliest_sample = r[0].t
+                r = r._raw
+                r.extend(response)
+                response = r
+                if earliest_sample <= pytz.timezone(NY).localize(start):
+                    got_all = True
+                else:
+                    delta = timedelta(days=1) if granularity == "day" else \
+                        timedelta(minutes=1)
+                    curr = earliest_sample - delta
+            return response
+
+        def _clear_out_of_market_hours(df):
+            """
+            only interested in samples between 9:30, 16:00 NY time
+            """
+            return df.between_time("09:30", "16:00")
+
+        def _drop_early_samples(df):
+            """
+            samples from server don't start at 9:30 NY time
+            let's drop earliest samples
+            """
+            for i, b in df.iterrows():
+                if i.time() == dtime(9, 30):
+                    return df[i:]
+
+        def _resample(df):
+            """
+            samples returned with certain window size (1 day, 1 minute) user
+            may want to work with different window size (5min)
+            """
+
+            if granularity == 'minute':
+                sample_size = f"{compression}Min"
+            else:
+                sample_size = f"{compression}D"
+            df = df.resample(sample_size).agg(
+                collections.OrderedDict([
+                    ('open', 'first'),
+                    ('high', 'max'),
+                    ('low', 'min'),
+                    ('close', 'last'),
+                    ('volume', 'sum'),
+                ])
+            )
+            if granularity == 'minute':
+                return df.between_time("09:30", "16:00")
+            else:
+                return df
+
+        # def _back_to_aggs(df):
+        #     response = []
+        #     for i, v in df.iterrows():
+        #         response.append({
+        #             "o": v.open,
+        #             "h": v.high,
+        #             "l": v.low,
+        #             "c": v.close,
+        #             "v": v.volume,
+        #             "t": i.timestamp() * 1000,
+        #         })
+        #     return Aggs({"results": response})
+
+        if not start:
+            response = self.oapi.get_barset(dataname,
+                                            granularity,
+                                            limit=1000,
+                                            end=end)[dataname]._raw
+        else:
+            response = _iterate_api_calls()
+        for bar in response:
+            # Aggs are in milliseconds, we multiply by 1000 to
+            # change seconds to ms
+            bar['t'] *= 1000
+        response = Aggs({"results": response})
+
+        cdl = response.df
+        if granularity == 'minute':
+            cdl = _clear_out_of_market_hours(cdl)
+            cdl = _drop_early_samples(cdl)
+        if compression != 1:
+            response = _resample(cdl)
+        # response = _back_to_aggs(cdl)
+        else:
+            response = response.df
+        response = response.dropna()
+        response = response[~response.index.duplicated()]
+        return response
+
     def streaming_prices(self, dataname, tmout=None):
         q = queue.Queue()
         kwargs = {'q': q, 'dataname': dataname, 'tmout': tmout}
